/*
 * Created on: Jul 28, 2024
 */

#include <stdint.h>
#include <ch32v20x.h>

#include "led.h"

#include "adc.h"



#define AW20X_DIM           31  // initial global current setting

#define AW20X_COLS          2
#define AW20X_ROWS          12
#define AW20X_FADE_COUNT    (AW20X_ROWS * AW20X_COLS)


static const uint16_t pwm_cie_256in_1024out[] = {
    0,    0,    1,    1,    2,    2,    3,    3,    4,    4,    4,    5,    5,    6,    6,    7,
    7,    8,    8,    8,    9,    9,   10,   10,   11,   11,   12,   12,   13,   13,   14,   15,
   15,   16,   17,   17,   18,   19,   19,   20,   21,   22,   22,   23,   24,   25,   26,   27,
   28,   29,   30,   31,   32,   33,   34,   35,   36,   37,   38,   39,   40,   42,   43,   44,
   45,   47,   48,   50,   51,   52,   54,   55,   57,   58,   60,   61,   63,   65,   66,   68,
   70,   71,   73,   75,   77,   79,   81,   83,   84,   86,   88,   90,   93,   95,   97,   99,
  101,  103,  106,  108,  110,  113,  115,  118,  120,  123,  125,  128,  130,  133,  136,  138,
  141,  144,  147,  149,  152,  155,  158,  161,  164,  167,  171,  174,  177,  180,  183,  187,
  190,  194,  197,  200,  204,  208,  211,  215,  218,  222,  226,  230,  234,  237,  241,  245,
  249,  254,  258,  262,  266,  270,  275,  279,  283,  288,  292,  297,  301,  306,  311,  315,
  320,  325,  330,  335,  340,  345,  350,  355,  360,  365,  370,  376,  381,  386,  392,  397,
  403,  408,  414,  420,  425,  431,  437,  443,  449,  455,  461,  467,  473,  480,  486,  492,
  499,  505,  512,  518,  525,  532,  538,  545,  552,  559,  566,  573,  580,  587,  594,  601,
  609,  616,  624,  631,  639,  646,  654,  662,  669,  677,  685,  693,  701,  709,  717,  726,
  734,  742,  751,  759,  768,  776,  785,  794,  802,  811,  820,  829,  838,  847,  857,  866,
  875,  885,  894,  903,  913,  923,  932,  942,  952,  962,  972,  982,  992, 1002, 1013, 1023,
};



struct LedMatrix led;
uint8_t boeing[BOEING_COUNT];

AW20x awled;
static uint8_t awled_fade[AW20X_FADE_COUNT];

static uint8_t led_matrix_updated = 0;



// helper for LED programs
void use_brightest(uint8_t *dest, uint8_t *compare, uint8_t count)
{
    while (count--) {
        if (*dest < *compare) *dest = *compare;
        dest++;
        compare++;
    }
}

static uint16_t pwm_gamma(uint8_t in)
{
    return pwm_cie_256in_1024out[in];
}

void led_init()
{
    TIM_TimeBaseInitTypeDef timer ={0};
    TIM_OCInitTypeDef pwm = {0};

    uint8_t i;


    // reset LED values
    for (i = 0; i < sizeof(led.all); i++) {
        led.all[i] = 0;
    }
    boeing[0] = boeing[1] = boeing[2] = boeing[3] = 0;


    // configure matrix
    awled.fade = awled_fade;
    for (i = 0; i < AW20X_FADE_COUNT; i++) {
        awled_fade[i] = 0;
    }

    aw20x_init(&awled, AW20X_ADDR_GND << 1, 2, 12, AW20X_IMAX_13_3MA);
    aw20x_commit_dim_global(&awled, AW20X_DIM);
    aw20x_led_enable(&awled, 0, 23);



    // configure rear bottom blue LEDs
    timer.TIM_Period = (1 << 10) - 1;       // 10-bit
    timer.TIM_Prescaler = 0;
    timer.TIM_ClockDivision = TIM_CKD_DIV1;
    timer.TIM_CounterMode = TIM_CounterMode_Up;
    TIM_TimeBaseInit(BOEING_TIM, &timer);

    pwm.TIM_OCMode = TIM_OCMode_PWM1;
    pwm.TIM_OutputState = TIM_OutputState_Enable;
    pwm.TIM_Pulse = 0;
    pwm.TIM_OCPolarity = TIM_OCPolarity_High;
    TIM_OC1Init(BOEING_TIM, &pwm);
    TIM_OC2Init(BOEING_TIM, &pwm);
    TIM_OC3Init(BOEING_TIM, &pwm);
    TIM_OC4Init(BOEING_TIM, &pwm);

    TIM_CtrlPWMOutputs(BOEING_TIM, ENABLE);
    TIM_OC1PreloadConfig(BOEING_TIM, TIM_OCPreload_Disable);
    TIM_ARRPreloadConfig(BOEING_TIM, ENABLE);
    TIM_Cmd(BOEING_TIM, ENABLE);
}

void led_matrix_is_updated()
{
    led_matrix_updated = 1;
}

void led_matrix_send()
{
    int8_t i, j;

    // only render when there's something to render
    if (led_matrix_updated) {
        // convert led matrix
        j = 0;

        // first RGBs are on the second select, in GBR order
        j = 12;
        for (i = 0; i < 4; i++) {
            awled.fade[j++] = led.ind.rgb[i][2];
            awled.fade[j++] = led.ind.rgb[i][1];
            awled.fade[j++] = led.ind.rgb[i][0];
        }

        // individual LEDs are kinda random, dunno why
        awled.fade[ 3] = led.ind.led[0];
        awled.fade[ 6] = led.ind.led[1];
        awled.fade[ 8] = led.ind.led[2];

        // last RGB is last on the first column
        awled.fade[ 9] = led.ind.rgb[4][2];
        awled.fade[10] = led.ind.rgb[4][1];
        awled.fade[11] = led.ind.rgb[4][0];

        // time to send
        aw20x_commit_fade(&awled);
        led_matrix_updated = 0;
    }
}

void led_boeing_update()
{
    uint8_t i;
    uint16_t scale[BOEING_COUNT];

    uint8_t lsens = adc_get_lsens_coarse();
    uint8_t max = adc_get_brightness(lsens);

    // this isn't a matrix so we can just update whenever
    // but we need to scale to ambient light level
    for (i = 0; i < BOEING_COUNT; i++) {
        scale[i] = max * boeing[i];
        scale[i] >>= 6;
    }

    TIM_CtrlPWMOutputs(BOEING_TIM, DISABLE);

    BOEING_TIM->CH1CVR = pwm_gamma(boeing[1]);
    BOEING_TIM->CH2CVR = pwm_gamma(boeing[0]);
    BOEING_TIM->CH3CVR = pwm_gamma(boeing[3]);
    BOEING_TIM->CH4CVR = pwm_gamma(boeing[2]);

    TIM_CtrlPWMOutputs(BOEING_TIM, ENABLE);
}
